BINDING_HEADER_ACTIONPAD2   = "Action Pad2";

BINDING_NAME_ACTIONBUTTON13       = "Action Button 13";
BINDING_NAME_ACTIONBUTTON14       = "Action Button 14";
BINDING_NAME_ACTIONBUTTON15       = "Action Button 15";
BINDING_NAME_ACTIONBUTTON16       = "Action Button 16";
BINDING_NAME_ACTPADCONFIG         = "Configure Toggle";
BINDING_NAME_ACTPADTOGGLEMAINBAR  = "Show/Hide MainBar"
BINDING_NAME_ACTPADTOGGLEEXTRABAR = "Toggle AP ExtraBar";

local PRE_DBG = "";

local function debug(msg)
	if (DEFAULT_CHAT_FRAME ~= nil) then
		if (PRE_DBG ~= "") then
			DEFAULT_CHAT_FRAME:AddMessage(PRE_DBG);
		end
		DEFAULT_CHAT_FRAME:AddMessage(msg);
	else
		PRE_DBG = PRE_DBG .. msg .. "\n";
	end
end

ACTPAD2_SavedVars = {};

local function GetSavedVar(name,def)
	local key = GetRealmName() .. "_" .. UnitName("player") .. "_" .. name;
	
	if (ACTPAD2_SavedVars[key] ~= nil) then
		return ACTPAD2_SavedVars[key];
	end
	return def;
end

local function SetSavedVar(name,val)
	local key = GetRealmName() .. "_" .. UnitName("player") .. "_" .. name;
	ACTPAD2_SavedVars[key] = val;
end


-------------------------------------------------------------------------------------------
-- Global overrides
-- local old_ActionButtonDown = ActionButtonDown;
-- local old_ActionButtonUp = ActionButtonUp;
-- local old_SetOverrideBindingClick = SetOverrideBindingClick

local function SetGlobalOverrides()
-- 	ActionButtonDown = function () end;
-- 	ActionButtonUp = function() end;
--	SetOverrideBindingClick = function(self, prispec,key, childname, button)
		
--		debug("setting binding click on " .. "prispec: " .. tostring(prispec) .. " key: " .. key ..  " childname: " .. childname .. " button " .. button);
--		old_SetOverrideBindingClick(self,prispec,key,childname,button);
--	end
end


------------------------------------------------------------------------------------------------
-- Child Button Setup
local NUM_PAGES   = 6;
local NUM_BUTTONS = 16;

local RANGE_CHECK_UPDATE = .250;
local TOOL_TIP_UPDATE    = .250;
local REPEAT_UPDATE      = .250;
local COOLDOWN_UPDATE    = .250;

local MAIN_BAR_VIS = "main_bar_vis";
local EXTRA_BAR_VIS = "extra_bar_vis";


local function GetButtonXY(btn,i)
	local btn_x = 0;
	local btn_y = 0;
	
	if (i >= 1 and i <= 5) then
		btn_x = (i - 1) * btn:GetWidth();
		btn_y = 0;
	elseif (i >= 6 and i <= 9) then
		btn_x = (i - 6) * btn:GetWidth();
		if (i > 7) then btn_x = btn_x + btn:GetWidth(); end
		btn_y = -btn:GetHeight();
	elseif (i >= 10 and i <= 11) then
		if (i == 10) then btn_x = 0; end
		if (i == 11) then btn_x = 4 * btn:GetWidth(); end
		btn_y = -btn:GetHeight() * 2;
	elseif (i >= 12 and i <= 16) then
		btn_x = (i - 12) * btn:GetWidth();
		btn_y = -btn:GetHeight() * 3;
	end
	return btn_x,btn_y;
end

local function buttonCheckLoadImage(self)
	local button = SecureStateChild_GetEffectiveButton(self);
    local type = SecureButton_GetModifiedAttribute(self, "type", button, "");
	
    if (type == "action") then
		local action = SecureButton_GetModifiedAttribute(self, "action", button, nil);
		action = tonumber(action);
		if (action) then
			local texture = GetActionTexture(action);
			if (texture) then
--				debug("Hmm setting texture to " .. texture);
				self.texture:SetTexture(texture);
				self.texture:Show();
			else
				self.texture:Hide();
			end
		end
	end
end

local function buttonCheckHotKey(self)
	local button = SecureStateChild_GetEffectiveButton(self);
    local type = SecureButton_GetModifiedAttribute(self, "type", button, "");
	
    if (type == "action") then
		local action = SecureButton_GetModifiedAttribute(self, "action", button, nil);
		action = tonumber(action);
		if (action and self.page ~= nil) then
			action = action - ((self.page -1) * NUM_BUTTONS);
			if (action <= (NUM_BUTTONS * NUM_PAGES)) then
				local key1,key2 = GetBindingKey("ACTIONBUTTON" .. action);
				
				if (key1 ~= nil) then
					self:SetAttribute("bindings-p" .. (self.page - 1), key1);
					self.hotkey:SetText(GetBindingText(key1,"KEY_"));
				else
					self:SetAttribute("bindings-p" .. (self.page - 1), "");
					self.hotkey:SetText("");
				end
			end
		end
	end
end

local function buttonCheckGrid(self)
	if (self.showgrid == 1) then
		if (not self.grid:IsVisible()) then
			self.grid:Show();
		end
	elseif (self.showgrid == 0) then
		if (self.grid:IsVisible()) then
			self.grid:Hide();
		end
	end
end

local function button_CheckRange(btn)
	local actionID = SecureButton_GetModifiedAttribute(btn, "action", button, nil);
	
	btn.hotkey:SetTextColor(.6,.6,.6);
	if (HasAction(actionID)) then
		if (ActionHasRange(actionID)) then
			local rangeFlag = IsActionInRange(actionID);
			
			if (rangeFlag == 1) then
				btn.hotkey:SetTextColor(0,1,0);
			elseif (rangeFlag == 0) then
				btn.hotkey:SetTextColor(1,0,0);
			end
		end
	end
end

local function button_CheckToolTip(btn)
	if (btn.toolTipTimeout ~= nil) then
		if (GameTooltip:IsOwned(btn:GetParent()) and btn:GetParent().toolTipOwner == btn) then
			if ( btn.toolTipTimeout > 0) then
				btn.toolTipTimeout = btn.toolTipTimeout - 1;
			else
				GameTooltip:Hide();
				btn.toolTipTimeout = nil;
				btn:GetParent().toolTipOwner = nil;
			end
		else
			btn.toolTipTimeout = 0;
		end
	end
	local frameID = GetMouseFocus();
	
	if (frameID ~= nil) then
		local mouseFrameName = frameID:GetName();
		if (mouseFrameName == btn:GetName()) then
			btn.toolTipTimeout = 1;
			GameTooltip:SetOwner(btn:GetParent(), "ANCHOR_BOTTOMRIGHT");
			btn:GetParent().toolTipOwner = btn;
			GameTooltip:SetAction(SecureButton_GetModifiedAttribute(btn, "action", button, nil));
		end
	end
end

local function button_RepeatUpdate(btn)
	local actionID = SecureButton_GetModifiedAttribute(btn, "action");
	
	if (IsCurrentAction(actionID) or IsAutoRepeatAction(actionID)) then
		if (IsAttackAction(actionID)) then
			if (btn.reptex:IsVisible()) then
				btn.reptex:Hide();
			else
				btn.reptex:Show();
			end
		else
			btn.grid:Show();
		end
	else
		if (btn.reptex:IsVisible()) then
			btn.reptex:Hide();
		end
		if (btn.grid:IsVisible() and (btn.showgrid or 0) == 0) then
			btn.grid:Hide();
		end
	end
end

local function button_CooldownUpdate(btn)
	local actionID = SecureButton_GetModifiedAttribute(btn, "action");
	local start, duration, enable = GetActionCooldown(actionID);
	if (start > 0) then
		if (btn.texture:GetAlpha() < 1) then
			btn.texture:SetAlpha(1);
		end
		
		local now = GetTime();
		local left = now - start;
		left = duration - left;
		if (left > 0) then
			local i,r = math.modf(left/60)
			r = r * 60;
			btn.cooldown:Show();
			btn.cooldownbk:Show();
			if (i > 0) then
				btn.cooldown:SetTextHeight(10);
				btn.cooldown:SetText(string.format("%d:%02d",i,r));
			else
				btn.cooldown:SetTextHeight(20);
				if (r < 1) then
					if (left < COOLDOWN_UPDATE) then
						btn.lastCooldown = COOLDOWN_UPDATE - left; -- make the end of the cooldown exact.
					end
					btn.cooldown:SetText("*");
				else
					btn.cooldown:SetText(string.format("%d",r));
				end
			end
		else
			btn.cooldown:Hide();
			btn.cooldownbk:Hide();
		end
	else
		if (btn.cooldown:IsVisible()) then
			btn.cooldown:Hide();
			btn.cooldownbk:Hide();
		end
		if (HasAction(actionID)) then
			isUsable, notEnoughMana = IsUsableAction(actionID)
			if (isUsable) then
				if (btn.texture:GetAlpha() < 1) then
					btn.texture:SetAlpha(1);
				end
			else
				if (btn.texture:GetAlpha() > .2) then
					btn.texture:SetAlpha(.2);
				end
			end
		end
	end
end

local function buttonUpdate(self, elapsed)
	if (not self.loaded) then
		self.loaded = true
		buttonCheckLoadImage(self);
		buttonCheckHotKey(self);
	end
	
	if (self.showgrid == nil) then
		buttonCheckGrid(self);
		self.showgrid = nil;
	end
	
	self.lastRangeCheck = self.lastRangeCheck + elapsed;
	if (self.lastRangeCheck >= RANGE_CHECK_UPDATE) then
		self.lastRangeCheck = 0;
		button_CheckRange(self);
	end
	
	self.lastToolTipCheck = self.lastToolTipCheck + elapsed;
	if (self.lastToolTipCheck >= TOOL_TIP_UPDATE) then
		self.lastToolTipCheck = 0;
		button_CheckToolTip(self);
	end
	
	self.lastRepeat = self.lastRepeat + elapsed;
	if (self.lastRepeat >= REPEAT_UPDATE) then
		self.lastRepeat = 0;
		button_RepeatUpdate(self);
	end
	
	self.lastCooldown = self.lastCooldown + elapsed;
	if (self.lastCooldown >= COOLDOWN_UPDATE) then
		self.lastCooldown = 0;
		button_CooldownUpdate(self);
	end
end

local function buttonDragStart(btn)
	if (LOCK_ACTIONBAR ~= "1") then
		local mouseOver = GetMouseFocus();
		if (mouseOver:GetName() ~= nil and mouseOver:GetName() == btn:GetName()) then
			local actionID = SecureButton_GetModifiedAttribute(btn, "action");
			PickupAction(actionID);
			buttonCheckLoadImage(btn);
			if (btn.floatFor ~= nil and btn.floatFor:IsVisible()) then
				buttonCheckLoadImage(btn.floatFor);
			end
		end
	end
end

local function buttonDragStop(btn)
	if (LOCK_ACTIONBAR ~= "1") then
		local mouseOver = GetMouseFocus();
		if (mouseOver:GetName() ~= nil and mouseOver:GetName() == btn:GetName()) then
			local actionID = SecureButton_GetModifiedAttribute(btn, "action");
			PlaceAction(actionID);
			buttonCheckLoadImage(btn);
			if (btn.floatFor ~= nil and btn.floatFor:IsVisible()) then
				buttonCheckLoadImage(btn.floatFor);
			end
		end
	end
end

local BUTTONS = {}

local function createPageSwitchButtons(parent,pageNum)
	-- create our page switch buttons.. that you can't see
	local pageButtonName = "AP2_BTN_PG_" .. (pageNum-1);
	local page = BUTTONS[pageButtonName];
	if (page == nil) then
		page = CreateFrame("Button", pageButtonName, parent, "SecureActionButtonTemplate");
		page:SetWidth(1);
		page:SetHeight(1);
		page:SetPoint("TOPLEFT", nil, "TOPLEFT", 0, 0);
		parent:SetAttribute("addchild",page);
	end
	
	page:SetAttribute("showstates", "*");
	page:SetAttribute("newstate", (pageNum-1));
	local key1,key2 = GetBindingKey("ACTIONPAGE" .. pageNum);
	if (key1) then
		for z = 1, NUM_PAGES do
			page:SetAttribute("bindings-p" .. (z-1),key1);
		end
	end
end

local function CreateChildButtons(parent)
	for q = 1, NUM_PAGES do
		for i = 1,NUM_BUTTONS do
			local buttonName = "AP2_PG" .. q .. "_BTN_" .. i;
			
			local btn = BUTTONS[buttonName];
			-- avoid allocating lots of buttons on multiple calls to the proc
			if (btn == nil) then
				btn = CreateFrame("Button", buttonName , parent, "AP2_btn");
				BUTTONS[buttonName] = btn;
				local btn_x,btn_y = GetButtonXY(btn,i);
				btn:SetPoint("TOPLEFT",parent:GetName(),"TOPLEFT",btn_x,btn_y);
				parent:SetAttribute("addchild", btn);
			end
			
			local actID       = i + ((q-1) * NUM_BUTTONS);
			local actionKeyID = actID  - ((q-1) * NUM_BUTTONS);
			
			local key1,key2 = GetBindingKey("ACTIONBUTTON" .. actionKeyID);
			
			-- debug("Looking at binding for action button ACTIONBUTTON" .. actionKeyID .. " key1 " .. (key1 or "nil"));
			if (key1) then
				btn:SetAttribute("bindings-p" .. (q-1), key1);
			else
				btn:SetAttribute("bindings-p" .. (q-1),nil);
			end
			
			btn:SetAttribute("showstates",tostring(q-1));
			btn.loadShowStates = btn:GetAttribute("showstates");
			
			btn:SetAttribute("type","action");
			btn:SetAttribute("action",actID);
		    btn:SetScript("OnUpdate", buttonUpdate);
			btn:RegisterForDrag("LeftButton", "RightButton");
			btn:SetScript("OnDragStart", buttonDragStart);
			btn:SetScript("OnReceiveDrag", buttonDragStop);
			btn.loaded     = false;
			btn.grid       = nil;
			btn.texture    = getglobal(btn:GetName() .. "Icon");
			btn.hotkey     = getglobal(btn:GetName() .. "HotKey");
			btn.cooldown   = getglobal(btn:GetName() .. "Cooldown");
			btn.cooldownbk = getglobal(btn:GetName() .. "CooldownBK");
			btn.grid       = getglobal(btn:GetName() .. "Grid");
			btn.reptex     = getglobal(btn:GetName() .. "Repeat");
			btn.count      = getglobal(btn:GetName() .. "Count");
			btn.page       = q;
			
			if (GetActionCount(actID) > 0) then
				btn.count:SetText(tostring(GetActionCount(actID)));
			end
			-- Update timers
			btn.lastToolTipCheck = 0;
			btn.lastRangeCheck   = 0;
			btn.lastRepeat       = 0;
			btn.lastCooldown     = 0;
		end
		
--		createPageSwitchButtons(parent,q);
	end
end

--
-- Patch the button visibility from a state, into another, and the key bindings
--
local function patchButtonsToState(key,fromState,toState,bindingTag)
	local buttonPage = fromState + 1;
	
	for i = 1,NUM_BUTTONS do
		local buttonName = "AP2_PG" .. buttonPage .. "_BTN_" .. i;
		local btn = BUTTONS[buttonName];
		
		btn:SetAttribute("showstates", btn:GetAttribute("showstates") .. "," .. toState);
		if (btn:GetAttribute("bindings-p" .. fromState) ~= nil and HasAction(btn:GetAttribute("action"))) then
--			btn:SetAttribute("bindings-" .. bindingTag, string.lower(key) .. "-" .. btn:GetAttribute("bindings-p" .. fromState));
			btn:SetAttribute("bindings-" .. bindingTag,btn:GetAttribute("bindings-p" .. fromState));
		end
	end
end

local TOGGLE_STATE_START = 6;

local TOGGLE_STATE_VAR = "togState";
local TOGGLE_STATE_DEF = { };

local function createToggleStates(header)
	-- make this handle being called after initialization
	--
	local TOGGLE_STATES = GetSavedVar(TOGGLE_STATE_VAR,TOGGLE_STATE_DEF);
	
	local curToggleState = TOGGLE_STATE_START;
	
	-- reset button visiblity to the page they were supposed to be on
	for _,btn in pairs(BUTTONS) do
		if (btn.loadShowStates ~= nil) then
			btn:SetAttribute("showstates", btn.loadShowStates);
			
			for i = 1,NUM_PAGES do
				btn:SetAttribute("bindings-SHIFT" ..i,nil);
				btn:SetAttribute("bindings-CTRL" ..i,nil);
				btn:SetAttribute("bindings-ALT" ..i,nil);
			end
		end
	end
	
	header:SetAttribute("statemap-SHIFT-1",nil);
	header:SetAttribute("statemap-SHIFT-0",nil);
	header:SetAttribute("statemap-CTRL-1",nil);
	header:SetAttribute("statemap-CTRL-0",nil);
	header:SetAttribute("statemap-ALT-1",nil);
	header:SetAttribute("statemap-ALT-0",nil);

	header.toggleStateBindings = "";
	
	for key,state in pairs(TOGGLE_STATES) do
		if ( (key == "SHIFT" or key =="CTRL" or key == "ALT") and (state ~= nil and state ~= "") ) then
			local toggleStates = "";

			local statemap = "";
			local upstatemap = "";
			
			for i = 0, (NUM_PAGES-1) do
			-- will make a state string like  6:SHIFT0;7:SHIFT1 ..etc
				toggleStates = toggleStates .. curToggleState .. ":"  .. key .. i .. ";";
				patchButtonsToState(key,state,curToggleState,key .. i);
				
				if (i ~= state) then
					statemap = statemap .. i .. ":" .. curToggleState .. ";";
					upstatemap = upstatemap .. curToggleState .. ":" .. i .. ";";
					header:SetAttribute("__page" .. curToggleState, state);
				end
				curToggleState = curToggleState + 1;
			end
			
			toggleStates = string.sub(toggleStates,0,-2);
			
			header.toggleStateBindings = header.toggleStateBindings .. ";" .. toggleStates;
			header:SetAttribute("statemap-" .. key .. "-1", statemap);
			header:SetAttribute("statemap-" .. key .. "-0", upstatemap);
		end
	end
	header:SetAttribute("statebindings",header.loadStateBindings .. header.toggleStateBindings);
end

local FORM_STATES_VAR = "formstates";
local FORM_STATES_DEF = { };

local function createFormStates(header)
	local FORM_STATES = GetSavedVar(FORM_STATES_VAR,FORM_STATES_DEF);

	for i = 1, GetNumShapeshiftForms() do
		header:SetAttribute("statemap-stance-" .. i, nil);
	end
	
	for form,state in pairs(FORM_STATES) do
		if (state ~= nil and state ~= "") then
--			debug("Setting form " .. form .. " to state " .. state)
			header:SetAttribute("statemap-stance-" .. form, "*:" .. state);
		end
	end
end

local function createPageChangeStates(header)
	for i = 1,NUM_PAGES do
		header:SetAttribute("statemap-actionbar-" .. i, (i-1))
	end
end


local DEF_FLOAT = { [0] = 0, [1] = 0, [2] = 0, [3] = 0, [4] = 0, [5] = 0 };
local FLOAT_VAR = "float_tab";

local FLOAT_FRAMES = {};

local function setupFloatButtons(floatFrame, page)
	local actStart = (page-1) * NUM_BUTTONS;
	for i = 1, NUM_BUTTONS do
		local btnName = floatFrame:GetName() .. "_btn_" .. i;
		
		btn = getglobal(btnName);
		BUTTONS[btnName] = btn;
		
		local actID = actStart + i;
		
		btn:SetAttribute("type","action");
		btn:SetAttribute("action",actID);
	    btn:SetScript("OnUpdate", buttonUpdate);
		btn:RegisterForDrag("LeftButton", "RightButton");
		btn:SetScript("OnDragStart", buttonDragStart);
		btn:SetScript("OnReceiveDrag", buttonDragStop);
		
		btn.loaded     = false;
		btn.grid       = nil;
		btn.texture    = getglobal(btn:GetName() .. "Icon");
		btn.hotkey     = getglobal(btn:GetName() .. "HotKey");
		btn.cooldown   = getglobal(btn:GetName() .. "Cooldown");
		btn.cooldownbk = getglobal(btn:GetName() .. "CooldownBK");
		btn.grid       = getglobal(btn:GetName() .. "Grid");
		btn.reptex     = getglobal(btn:GetName() .. "Repeat");
		btn.count      = getglobal(btn:GetName() .. "Count");
		
		local padButtonName = "AP2_PG" .. page .. "_BTN_" .. i;
		btn.floatFor   = BUTTONS[padButtonName];
			
		if (GetActionCount(actID) > 0) then
			btn.count:SetText(tostring(GetActionCount(actID)));
		end
		-- Update timers
		btn.lastToolTipCheck = 0;
		btn.lastRangeCheck   = 0;
		btn.lastRepeat       = 0;
		btn.lastCooldown     = 0;
	end
end

local function createFloats(header)
	local FLOATS = GetSavedVar(FLOAT_VAR,DEF_FLOAT);
	
	for pad,onOff in pairs(FLOATS) do
		local floatName = "AP2_float_" .. pad;
		local floatFrame = FLOAT_FRAMES[floatName];
		if (floatFrame == nil and onOff == 1) then
			floatFrame = CreateFrame("Frame", floatName , UIParent, "AP2_float_template");
			FLOAT_FRAMES[floatName] = floatFrame;
			floatFrame.bank = getglobal(floatFrame:GetName() .. "_BANK");
			floatFrame.bank:SetText("Pad" .. tostring(pad+1));
			floatFrame:SetClampedToScreen(1);
			local posX = GetSavedVar(floatName .. "posX", 0);
			local posY = GetSavedVar(floatName .. "posY", 0);
			if (posY == 0 and posX == 0) then
				floatFrame:SetPoint("CENTER");
			else
				floatFrame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", posX, posY);
			end
			floatFrame:SetScale(GetSavedVar(floatName .. "_scale",1));
			setupFloatButtons(floatFrame,pad+1);
		end

		if (onOff == 1) then
			if (not floatFrame:IsVisible()) then
				floatFrame:Show();
			end
		else
			if (floatFrame ~= nil and floatFrame:IsVisible()) then
				floatFrame:Hide();
			end
		end
	end
end

-- save out the floating frame's position for load
function AP2_savefloat_pos(self)
	local point,relativeTo,relativePoint,xOfs,yOfs = self:GetPoint()
	local Xvar = self:GetName() .. "posX";
	local Yvar = self:GetName() .. "posY";
	SetSavedVar(Xvar, xOfs );
	SetSavedVar(Yvar, yOfs );
end

local function DoLoad(self)
	self:SetAttribute("statebindings","");
	CreateChildButtons(self);
	self.bank = getglobal(self:GetName() .. "_BANK");
	
	if (GetSavedVar(MAIN_BAR_VIS,0) == 1) then
		if (not MainMenuBar:IsVisible()) then
			ShowUIPanel(MainMenuBar);
		end
	else
		if (MainMenuBar:IsVisible()) then
			HideUIPanel(MainMenuBar);
		end
	end

	-- prepend our default states on to the button, toggles may have set this when setting up the child buttons.

--	hooksecurefunc("SecureStateHeader_OnAttributeChanged",
--		function (self,name,value)
--			debug("Attrib name: " .. name);
--		end);
	self:SetScale(GetSavedVar(self:GetName() .. "_scale",1));

		
	self:SetAttribute("statebindings", "0:p0;1:p1;2:p2;3:p3;4:p4;5:p5" .. self:GetAttribute("statebindings"));
	self.loadStateBindings = self:GetAttribute("statebindings");
	
	createToggleStates(self);
	createFormStates(self);
	createPageChangeStates(self);
	createFloats(self);
	
	self.StateChanged =
		function (self,value)
			AP2_pad_cfg_Setup_drdToggle_StateChange();
			
			value = tonumber(value);
			if (value >= NUM_PAGES) then
				value = self:GetAttribute("__page" .. value);
			end

			
			self.bank:SetText( (tonumber(value)+1) .. "/" .. NUM_PAGES);
		end
	-- force a binding refresh basiclly
	local state = self:GetAttribute("state") or 0;
	local swap = state + 1
	if (swap > NUM_PAGES) then
		swap = 0
	end
	self:SetAttribute("state",swap);
	self:SetAttribute("state",state);
		
--	debug("/ap2 help , to see help");
end

local function button_UpdateCounts()
	for k,btn in pairs(BUTTONS) do
		local actionID = SecureButton_GetModifiedAttribute(btn, "action");
		if (IsConsumableAction(actionID)) then
			btn.count:SetText(tostring(GetActionCount(actionID)));
		else
			btn.count:SetText("");
		end
	end
end

local function button_CheckLoadImages()
	for k,btn in pairs(BUTTONS) do
		buttonCheckLoadImage(btn);
	end
end

local function button_ShowGrids(self)
	for k,btn in pairs(BUTTONS) do
		btn.grid:Show();
		btn.showgrid = 1;
	end
end

local function button_HideGrids(self)
	for k,btn in pairs(BUTTONS) do
		btn.grid:Hide();
		btn.showgrid = 0;
	end
end

local function AP2_SlashCmd(cmd)
	if (cmd == "cfg") then
		AP2_Toggle_Configure();
	else
		debug("/ap2 cfg  ,to toggle the configuration mode.");
		debug("/ap2 help ,to see this help.");
	end
end

local AP2_HDR = nil;
local AP2_Pending_Load = nil;

local function CheckPendingLoad()
	if (InCombatLockdown()) then
		AP2_Pending_Load = true
--		debug("Lockdown stalling")
	elseif (AP2_Pending_Load) then
		AP2_Pending_Load = false
		DoLoad(AP2_HDR)
--		debug("Non lockdown Reloading")
	end
end

-----------------------------------------------------------------------------
-- Setup our Header that will manage the states of the buttons
function AP2_hdr_Load(self)
	AP2_HDR = self;
	
	AP2_Pending_Load = true;
	
	SlashCmdList["AP2"] = AP2_SlashCmd;
	SLASH_AP21 = "/actionpad2" 
	SLASH_AP22 = "/ap2" 

	
	SetGlobalOverrides()
	MainMenuBar:Hide()
	
	self:SetClampedToScreen(1)
	
	SecureStateDriver_OnLoad(self)
	
	self:RegisterEvent("UPDATE_BINDINGS")
	self:RegisterEvent("ACTIONBAR_SLOT_CHANGED")
	self:RegisterEvent("ACTIONBAR_SHOWGRID")
	self:RegisterEvent("ACTIONBAR_HIDEGRID")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("ADDON_LOADED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")

	hooksecurefunc("SecureStateDriver_OnEvent",
		function (self,event,...)
			if (self == AP2_HDR) then
				if (event == "UPDATE_BINDINGS") then
					AP2_Pending_Load = true;
					CheckPendingLoad();
				elseif (event == "ACTIONBAR_SLOT_CHANGED") then
--					button_CheckLoadImages();
					button_UpdateCounts();
					AP2_Pending_Load = true;
					CheckPendingLoad();
				elseif (event == "ACTIONBAR_SHOWGRID") then
					button_ShowGrids(self);
				elseif (event == "ACTIONBAR_HIDEGRID") then
					button_HideGrids(self);
				elseif (event == "PLAYER_ENTERING_WORLD") then
					AP2_Pending_Load = true;
					CheckPendingLoad();
				elseif (event == "ADDON_LOADED" and select(1,...) == "ActionPad2") then
					debug("/ap2 cfg - to toggle the configuration mode.");
				elseif (event == "PLAYER_REGEN_ENABLED") then
					CheckPendingLoad();
				end
			end
		end);
end


local NUM_EXTRA_BUTTONS = 24
-----------------------------------------------------------------------
-- Extra Pad area
local function GetExtraButtonXY(btn,i)
	local i,r = math.modf( (i-1)/4);
	local xpos = r * 4;
	if (xpos > 1) then xpos = xpos + 1; end
	
	xpos = xpos * btn:GetWidth();
	
	local ypos = i * -btn:GetHeight();
	return xpos,ypos;
end

local function CreateExtraChildButtons(parent)
	for i = 1,NUM_EXTRA_BUTTONS do
		local buttonName = "AP2_EXTRA_BTN_" .. i;
		
		local btn = BUTTONS[buttonName];
		-- avoid allocating lots of buttons on multiple calls to the proc
		if (btn == nil) then
			btn = CreateFrame("Button", buttonName , parent, "AP2_btn");
			BUTTONS[buttonName] = btn;
			local btn_x,btn_y = GetExtraButtonXY(btn,i);
			btn:SetPoint("TOPLEFT",parent:GetName(),"TOPLEFT",btn_x,btn_y);
			parent:SetAttribute("addchild", btn);
		end
		
		local actID = (NUM_PAGES * NUM_BUTTONS) + i;
		
		btn:SetAttribute("showstates","*");
		btn:SetAttribute("type","action");
		btn:SetAttribute("action",actID);
		btn:SetScript("OnUpdate", buttonUpdate);
		btn:RegisterForDrag("LeftButton", "RightButton");
		btn:SetScript("OnDragStart", buttonDragStart);
		btn:SetScript("OnReceiveDrag", buttonDragStop);
		btn.loaded = false;
		btn.grid   = nil;
		btn.texture    = getglobal(btn:GetName() .. "Icon");
		btn.hotkey     = getglobal(btn:GetName() .. "HotKey");
		btn.cooldown   = getglobal(btn:GetName() .. "Cooldown");
		btn.cooldownbk = getglobal(btn:GetName() .. "CooldownBK");
		btn.grid       = getglobal(btn:GetName() .. "Grid");
		btn.reptex     = getglobal(btn:GetName() .. "Repeat");
		btn.count      = getglobal(btn:GetName() .. "Count");
		
		if (GetActionCount(actID) > 0) then
			btn.count:SetText(tostring(GetActionCount(actID)));
		end
		-- Update timers
		btn.lastToolTipCheck = 0;
		btn.lastRangeCheck   = 0;
		btn.lastRepeat       = 0;
		btn.lastCooldown     = 0;
	end
end

local function DoExtraLoad(self)
	CreateExtraChildButtons(self);
	self:SetScale(GetSavedVar(self:GetName() .. "_scale",1));
	
	if (GetSavedVar(EXTRA_BAR_VIS,1) == 1) then
		if (not self:IsVisible()) then
			self:Show();
		end
	else
		if (self:IsVisible()) then
			self:Hide();
		end
	end
end

function AP2_hdr_LoadExtra(self)
	self:RegisterEvent("UPDATE_BINDINGS");
	self:RegisterEvent("ACTIONBAR_SLOT_CHANGED");
	self:RegisterEvent("ACTIONBAR_SHOWGRID");
	self:RegisterEvent("ACTIONBAR_HIDEGRID");
	self:SetClampedToScreen(1);
	
	self:SetScript("OnEvent",
		function(self,event,arg1)
			if (event == "UPDATE_BINDINGS") then
				DoExtraLoad(self);
			elseif (event == "ACTIONBAR_SLOT_CHANGED") then
--				button_CheckLoadImages();
--				button_UpdateCounts();
			elseif (event == "ACTIONBAR_SHOWGRID") then
--				button_ShowGrids();
			elseif (event == "ACTIONBAR_HIDEGRID") then
--				button_HideGrids();
			end
		end);
end

------------------------------------------------------------------------------------------
---       Configuration
------------------------------------------------------------------------------------------
local CONFIG_FRAMES = { ["pad"] = nil,
						["padextra"] = nil };

local function pad_config_onUpdate(self,elapsed)
	local newScale = self.slider:GetValue();
	local curScale = self.control_parent:GetScale();
	if (newScale ~= curScale) then
		self.slider.text:SetText("Scale " .. newScale ..  "%");
		local scale = newScale / 100;
		self.control_parent:SetScale(scale);
		self.control_parent:SetPoint("TOPLEFT",self:GetLeft() * (1/scale), (-(self:GetParent():GetTop() - self:GetTop()) - self:GetHeight()) * (1/scale));
		
		SetSavedVar(self.control_parent:GetName() .. "_scale",scale);
	end
end

local function createConfigPad(name, control_parent)
	local cfg_pad = CreateFrame("Frame",name, UIParent, nil);
	cfg_pad.control_parent = control_parent;
	cfg_pad:SetWidth(250);
	cfg_pad:SetHeight(70);
	local bg = cfg_pad:CreateTexture(nil, "BACKGROUND");
	bg:SetTexture(0.5,0.5,0.5,0.5);
	bg:SetAllPoints(cfg_pad);
	
	cfg_pad:SetScript("OnUpdate",pad_config_onUpdate);
--	cfg_pad:SetPoint("TOPLEFT",UIParent,"TOPLEFT",control_parent:GetLeft(),-(cfg_pad:GetParent():GetTop() - control_parent:GetTop()) + cfg_pad:GetHeight());
	cfg_pad:Hide();

	
	local lbl = cfg_pad:CreateFontString(nil, "TEXT");
	lbl:SetFontObject(GameFontNormalSmall);
	lbl:SetPoint("BOTTOMLEFT", cfg_pad, "TOPLEFT", 5, -10);
	lbl:SetText("Configuration");
	
	cfg_pad.slider = CreateFrame("Slider",name.."_scale", cfg_pad, "OptionsSliderTemplate");
	cfg_pad.slider:SetPoint("TOPLEFT",cfg_pad,"TOPLEFT",25,-20);
	cfg_pad.slider:SetWidth(cfg_pad:GetWidth() - 50);
	cfg_pad.slider:SetOrientation("HORIZONTAL");
	cfg_pad.slider:SetMinMaxValues(50,200);
	cfg_pad.slider:SetValue(control_parent:GetScale() * 100);
	cfg_pad.slider:SetValueStep(10);
	cfg_pad.slider.text = getglobal(cfg_pad.slider:GetName() .. "Text");
	cfg_pad.slider.text:SetText("Scale " .. cfg_pad.slider:GetValue() ..  "%");
	return cfg_pad;
end

local function checkLoadToggle(key,cfg_name,config_parent,onOff)
	local cfg_pad = CONFIG_FRAMES[key];
	if (cfg_pad == nil) then
		cfg_pad = createConfigPad(cfg_name, config_parent);
		CONFIG_FRAMES[key] = cfg_pad;
	end
	
	if (onOff == 1) then
		if (config_parent:IsVisible()) then
			local pX = cfg_pad.control_parent:GetLeft() * cfg_pad.control_parent:GetScale();
			local pY = cfg_pad.control_parent:GetScale() * cfg_pad.control_parent:GetTop();
			cfg_pad:SetPoint("TOPLEFT",UIParent,"TOPLEFT", pX, -(cfg_pad:GetParent():GetTop() - pY)  + cfg_pad:GetHeight());
			cfg_pad:Show();
		else
			cfg_pad:Hide();
		end
	else
		cfg_pad:Hide();
	end
end

local function checkFloatToggle(padCfg)
	local toggleBtn = getglobal(padCfg:GetName() .. "_phloat");
	local FLOAT_STATES = GetSavedVar(FLOAT_VAR,DEF_FLOAT);
	if (FLOAT_STATES[tonumber(AP2_HDR:GetAttribute("state"))] > 0) then
		toggleBtn:SetChecked(1);
	else
		toggleBtn:SetChecked(0);
	end
end

local function checkLoadPadConfig(positionto)
	local padCfg = getglobal("AP2_pad_cfg");
	if (padCfg:IsVisible()) then
		padCfg:Hide();
	else
		padCfg:SetPoint("TOPLEFT",positionto,"TOPRIGHT",0,0);
		padCfg:Show();
		
		checkFloatToggle(padCfg);
	end
end

function checkLoadFloatToggles(force)
	-- make scaling configs for thefloats
	for _, floatFrame in pairs(FLOAT_FRAMES) do
		local name = "padfloat_" .. floatFrame:GetName();
		checkLoadToggle(name,name,floatFrame,force);
	end
end

local configOn = 0;

function AP2_Toggle_Configure()
	if (configOn == 0) then
		configOn = 1;
	else
		configOn = 0;
	end
	checkLoadToggle("pad","pad_cfg", AP2_HDR, configOn);
	checkLoadToggle("padextra","extra_cfg", getglobal("AP2_hdr_extra"), configOn);
	checkLoadFloatToggles(configOn);
	checkLoadPadConfig(AP2_HDR);
end

local function findToggleForState(curState)
	local TOGGLE_STATES = GetSavedVar(TOGGLE_STATE_VAR,TOGGLE_STATE_DEF);
	for modifier,state in pairs(TOGGLE_STATES) do
		if (state == curState) then
			return modifier;
		end
	end
	return nil;
end

--
-- Ugh annoying.. ok the global "this" variable is the button that has a copy of the properties set from the "info" table created when we called UIDropDownMenu_AddButton
--
local function ap2_pad_cfg_Setup_drdToggle_OnClick(arg1,arg2)
	local TOGGLE_STATES = GetSavedVar(TOGGLE_STATE_VAR,TOGGLE_STATE_DEF);
	local curstate = tonumber(AP2_HDR:GetAttribute("state"));
	
	--- clear out all settings that have my current state
	for k,v in pairs(TOGGLE_STATES) do
		if (v == curstate) then
			TOGGLE_STATES[k] = "";
		end
	end

	-- If we have a value, meaning we selected a modifier option in the dropdown
	if (this.value ~= nil) then
		TOGGLE_STATES[this.value] = curstate; -- set the modifier to the current state.... so hitting that modifer will go back to the state we are in now.
	end
	
	SetSavedVar(TOGGLE_STATE_VAR,TOGGLE_STATES);
	createToggleStates(AP2_HDR);
	
	UIDropDownMenu_SetSelectedValue(this.owner, this.value, 0);
end

local function ap2_pad_cfg_Setup_drdToggle_Init()
	local MODIFIERS = { "SHIFT", "CTRL", "ALT" };
	local info = UIDropDownMenu_CreateInfo();
	local owner = getglobal("AP2_pad_cfg_drdToggle");

	-- create our None Option
	info.text  = "None";
	info.value = nil;
	info.func  = ap2_pad_cfg_Setup_drdToggle_OnClick;
	info.owner = owner;
	local curState = findToggleForState(tonumber(AP2_HDR:GetAttribute("state")));
	info.check = (curState == nil);
	if (info.check) then
		UIDropDownMenu_SetText(info.text,info.owner);
	end
	UIDropDownMenu_AddButton(info);

	
	for _,v in pairs(MODIFIERS) do
		info = UIDropDownMenu_CreateInfo();
		info.text    = v;
		info.value   = v;
		info.func    = ap2_pad_cfg_Setup_drdToggle_OnClick;
		info.owner   = owner;
		
		if (curState == v) then
			info.checked = 1;
			UIDropDownMenu_SetText(info.text,info.owner);
		else
			info.checked = nil;
		end
		
		UIDropDownMenu_AddButton(info);
	end
end

function AP2_pad_cfg_Setup_drdToggle_StateChange()
	local padCfg = getglobal("AP2_pad_cfg");
	if (padCfg:IsVisible()) then
		-- re init if we get a state change while the config is up.
		padCfg:Hide();
		padCfg:Show();
		checkFloatToggle(padCfg);
	end
end

-----------------------------------------------------------------------
-- Form states

local function findFormForState(formstate)
	local FORM_STATES = GetSavedVar(FORM_STATES_VAR,FORM_STATES_DEF);
	for form,state in pairs(FORM_STATES) do
		if (state == formstate) then
			return form;
		end
	end
	return nil;
end


function AP2_pad_cfg_Setup_drdToggle(self)
	UIDropDownMenu_Initialize(self, ap2_pad_cfg_Setup_drdToggle_Init);
	UIDropDownMenu_SetSelectedValue(self, findToggleForState(tonumber(AP2_HDR:GetAttribute("state"))), 1);
	UIDropDownMenu_SetWidth(90, self);
end


local function ap2_pad_cfg_Setup_drdForm_OnClick(arg1,arg2)
	local FORM_STATES = GetSavedVar(FORM_STATES_VAR,FORM_STATES_DEF);
	local curstate = tonumber(AP2_HDR:GetAttribute("state"));
	
	--- clear out all settings that have my current state
	for form,state in pairs(FORM_STATES) do
		if (state == curstate) then
			FORM_STATES[form] = "";
		end
	end

	-- If we have a value, meaning we selected a modifier option in the dropdown
	if (this.value ~= nil) then
		FORM_STATES[this.value] = curstate; -- set the modifier to the current state.... so hitting that modifer will go back to the state we are in now.
	end
	
	SetSavedVar(FORM_STATES_VAR,FORM_STATES);
	createFormStates(AP2_HDR);
	
	UIDropDownMenu_SetSelectedValue(this.owner, this.value);
end

local function ap2_pad_cfg_Setup_drdForm_Init()
	local info = UIDropDownMenu_CreateInfo();
	local owner = getglobal("AP2_pad_cfg_drdForm");

	-- create our None Option
	info.text  = "None";
	info.value = nil;
	info.func  = ap2_pad_cfg_Setup_drdForm_OnClick;
	info.owner = owner;
	local curForm = findFormForState(tonumber(AP2_HDR:GetAttribute("state")));
	info.check = (curForm == nil);
	UIDropDownMenu_SetText(info.text,info.owner);
	UIDropDownMenu_AddButton(info);
	
	if (GetNumShapeshiftForms() > 0) then
		info = UIDropDownMenu_CreateInfo();
		info.text  = "Unformed";
		info.value = 0;
		info.func  = ap2_pad_cfg_Setup_drdForm_OnClick;
		info.owner = owner;
		if (curForm == 0) then
			info.checked = 1;
			UIDropDownMenu_SetText(info.text,info.owner);
		else
			info.check = nil;
		end
		UIDropDownMenu_AddButton(info);

		
		for i=1,GetNumShapeshiftForms() do
			info = UIDropDownMenu_CreateInfo();
			local icon, name = GetShapeshiftFormInfo(i);

			info.text    = name;
			info.value   = i;
			info.func    = ap2_pad_cfg_Setup_drdForm_OnClick;
			info.owner   = owner;
			
			if (curForm == i) then
				info.checked = 1;
				UIDropDownMenu_SetText(info.text,info.owner);
			else
				info.checked = nil;
			end
			
			UIDropDownMenu_AddButton(info);
		end
	end
end

function AP2_pad_cfg_Setup_drdForm(self)
	UIDropDownMenu_Initialize(self, ap2_pad_cfg_Setup_drdForm_Init);
	UIDropDownMenu_SetSelectedValue(self, findFormForState(tonumber(AP2_HDR:GetAttribute("state"))));
	UIDropDownMenu_SetWidth(90, self);
end

function AP2_pad_cfg_Float(self)
	-- enable a float pad for the current state
	local save_var = GetSavedVar(FLOAT_VAR,DEF_FLOAT);
	local curstate = tonumber(AP2_HDR:GetAttribute("state"));
	save_var[curstate] = self:GetChecked() or 0;
	SetSavedVar(FLOAT_VAR,save_var);
	createFloats(AP2_HDR);
	checkLoadFloatToggles(self:GetChecked() or 0);
end


function AP2_Toggle_Mainbar()
	if (MainMenuBar:IsVisible()) then
		SetSavedVar(MAIN_BAR_VIS,0);
		HideUIPanel(MainMenuBar);
	else
		SetSavedVar(MAIN_BAR_VIS,1);
		ShowUIPanel(MainMenuBar);
	end
end

function AP2_Toggle_Extrabar()
	local extra = getglobal("AP2_hdr_extra");
	if (extra:IsVisible()) then
		SetSavedVar(EXTRA_BAR_VIS,0);
		extra:Hide();
	else
		SetSavedVar(EXTRA_BAR_VIS,1);
		extra:Show();
	end
end
